package algebra.linear;

import java.util.*;

import algebra.numbers.Rational;
import algebra.words.Alphabet;
import algebra.words.Word;

/**
 * @author Anton Mellit
 *
 * To change this generated comment edit the template variable "typecomment":
 * Window>Preferences>Java>Templates.
 * To enable and disable the creation of type comments go to
 * Window>Preferences>Java>Code Generation.
 */
public class Sum {
	protected SortedMap map;
	public Sum() {
		map = new TreeMap();
	}
	public void setZero() {
		map.clear();
	}
	public void add(Object o) {
		add(o, Rational.ONE);
	}
	public void add(Object o, long k) {
		add(o, Rational.valueOf(k));
	}
	public void add(Object o, Rational k) {
		if(!k.isZero()) {
			Rational x = (Rational) map.get(o);
			if(x!=null) {
				k = x.add(k);
				if(k.isZero()) {
					map.remove(o);
				} else {
					map.put(o, k);
				}
			} else {
				map.put(o, k);
			}
		}
	}
	public void add(Sum s) {
		add(s, Rational.ONE);
	}
	public void add(Sum s, long k) {
		add(s, Rational.valueOf(k));
	}
	public void add(Sum s, Rational k) {
		for(Iterator i = s.map.entrySet().iterator(); i.hasNext();) {
			Map.Entry entry = (Map.Entry) i.next();
			add(entry.getKey(), k.multiply((Rational) entry.getValue()));
		}
	}		
	public void multiply(Rational k) {
		if(k.isZero()) {
			map.clear();
		}
		for(Iterator i = map.entrySet().iterator(); i.hasNext();) {
			Map.Entry entry = (Map.Entry) i.next();
			map.put(entry.getKey(), k.multiply((Rational) entry.getValue()));
		}
	}
    public String toString() {
    	return toString(false);
    }
    public String toString(boolean skipMiddle) {
        if(map.entrySet().isEmpty()) return "0";
        boolean first = true;
        String s = "";
        int j = 0;
        boolean skipped = false;
        for (Iterator iterator = map.entrySet().iterator(); iterator.hasNext();) {
        	if(j<20 || map.size()-j<20 || !skipMiddle) {
	            Map.Entry entry = (Map.Entry) iterator.next();
    	        Object o = entry.getKey();
            
        	    Rational k = (Rational) entry.getValue();
        	    if(skipped) {
        	    	s = s + " ... ";
        	    }
            	if(k.compareTo(Rational.ZERO)>=0) {
	            	if(!first) {
    	        		s = s + " + ";
        	    	}
            	} else {
            		if(!first) {
			            s = s + " - ";
    	        	} else {
        	    		s = s + "- ";
            		}
	            }
    	        k = k.getAbs();
        	    String ostr = o.toString();
	        	if(!k.equals(Rational.ONE) || ostr.length()==0) {
                	 s = s + k;
	             }
    	        s = s + ostr;
        	    first = false;
            	skipped = false;
        	} else {
        		skipped = true;
        	}
           	j++;
        }
        return s;
    }
    public void normalize() {
        if(map.isEmpty()) return;
        Object w = map.lastKey();
        Rational k = (Rational) map.get(w);
        multiply(k.invert());
    }
    public Object getMainObject() {
    	if(map.isEmpty()) {
    		return null;
    	} else {
	    	return map.lastKey();
    	}
    }
    public Set getObjects() {
    	return map.keySet();
    }
    public Rational getKoeff(Object o) {
    	Rational k = (Rational) map.get(o);
    	if(k==null) {
    		return Rational.ZERO;
    	} else {
    		return k;
    	}
    }
    public boolean isZero() {
    	return map.isEmpty();
    }
    public boolean equals(Object o) {
    	if(o instanceof Sum) {
    		Sum s = (Sum) o;
    		Iterator i1 = map.entrySet().iterator();
    		Iterator i2 = s.map.entrySet().iterator();
    		while(i1.hasNext()) {
    			if(!i2.hasNext()) {
    				return false;
    			}
    			Map.Entry e1 = (Map.Entry) i1.next();
    			Map.Entry e2 = (Map.Entry) i2.next();
    			if(!e1.getKey().equals(e2.getKey())) {
    				return false;
    			}
    			if(!e1.getValue().equals(e2.getValue())) {
    				return false;
    			}
    		}
    		if(i2.hasNext()) {
    			return false;
    		}
    		return true;
    	} else {
    		return super.equals(o);
    	}
    }
    public void modulo(Sum s) {
    	Object main = s.getMainObject();
    	Rational k = ((Rational) map.get(main)).divide((Rational) s.map.get(main));
    	if(k!=null) {
    		add(s, k.multiply(-1));
    	}
    }
}
