package algebra.words;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * @author Anton Mellit
 *
 * To change this generated comment edit the template variable "typecomment":
 * Window>Preferences>Java>Templates.
 * To enable and disable the creation of type comments go to
 * Window>Preferences>Java>Code Generation.
 */
public class Word implements Comparable{
	private Letter[] letters;
	private int centrals;
	public static final Word ONE = new Word();
	
	public Word() {
		this.letters = new Letter[0];
		normalize();
	}
	public Word(Letter l) {
		this.letters = new Letter[1];
		this.letters[0] = l;
		normalize();
	}
	public Word(Letter[] letters) {
		this.letters = letters;
		normalize();
	}
	public void normalize() {
		boolean ok = false;
		centrals = 0;
		while(!ok) {
			ok = true;
			for(int i = 0; i<letters.length-1; i++) {
				if(letters[i+1].isCentral()) {
					if(!letters[i].isCentral() || 
						(letters[i].compareTo(letters[i+1])>0)) {
							ok = false;
							Letter z = letters[i];
							letters[i] = letters[i+1];
							letters[i+1] = z;
						}
				}
			}
		}
		for(int i = 0; i<letters.length; i++) {
			if(letters[i].isCentral()) {
				centrals++;
			} else {
				break;
			}
		}
	}
	
	public int getLength() {
		return letters.length;
	}
	
	public Word multiply(Word left, Word right) {
		int len = letters.length+left.letters.length+right.letters.length;
		Letter[] arr = new Letter[len];
		int j = 0;
		for(int i = 0; i<left.letters.length; i++) {
			arr[j] = left.letters[i];
			j++;
		}
		for(int i = 0; i<letters.length; i++) {
			arr[j] = letters[i];
			j++;
		}
		for(int i = 0; i<right.letters.length; i++) {
			arr[j] = right.letters[i];
			j++;
		}
		return new Word(arr);
	}
	
	protected int[] getNextLevel(int prevLevel) {
		int level = Integer.MIN_VALUE;
		int num = 0;
		for(int i = 0; i<letters.length; i++) {
			int l = letters[i].getLevel();
			if(l>level && l<prevLevel) {
				level = l;
				num = 1;
			} else if(l==level) {
				num++;
			}
		}
		int[] ans = {level, num};
		return ans;
	}
	
	protected int compareToLevel(Word w, int level) {
		int i1 = 0;
		int i2 = 0;
		while(true) {
			while(i1<letters.length && letters[i1].getLevel()<level) i1++;
			if(i1==letters.length) break;
			while(i2<w.letters.length && w.letters[i2].getLevel()<level) i2++;
			int c = letters[i1].compareTo(w.letters[i2]);
			if(c!=0) return c;
			i1++;
			i2++;
		}
		return 0;
	}
	
	/**
	 * @see java.lang.Comparable#compareTo(Object)
	 */
	public int compareTo(Object o) {
		Word w = (Word) o;
		int lastLevel = Integer.MAX_VALUE;
		while(lastLevel>Integer.MIN_VALUE) {
			int[] a1 = getNextLevel(lastLevel);
			int[] a2 = w.getNextLevel(lastLevel);
			if(a1[0]!=a2[0]) {
				if(a1[0]<a2[0]) {
					return -1;
				} else {
					return 1;
				}
			}
			lastLevel = a1[0];
			if(a1[1]!=a2[1]) {
				return a1[1]-a2[1];
			}
			int c = compareToLevel(w, lastLevel);
			if(c!=0) {
				return c;
			}
		}
		return 0;
	}
	
	/**
	 * @see java.lang.Object#equals(Object)
	 */
	public boolean equals(Object obj) {
		if(obj instanceof Word) {
			Word w = (Word) obj;
			return Arrays.equals(letters, w.letters);
		}
		return super.equals(obj);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
        if(letters.length==0) return "";
        String s = "";
        Letter lastLetter = null;
        int lastCount = 0;
        for(int i = 0; i<letters.length; i++) {
            Letter letter = letters[i];
            if(letter==lastLetter) {
                lastCount++;
            } else {
                if(lastCount!=0) {
                    s = s + lastLetter;
                    if(lastCount!=1) {
                        s = s + "^"+lastCount;
                    }
                }
                lastCount = 1;
                lastLetter = letter;
            }
        }
        s = s + lastLetter;
        if(lastCount!=1) {
            s = s + "^"+lastCount;
        }
        return s;
	}
	
	public Word subword(int begin, int end) {
		Letter[] l = new Letter[end-begin];
		for(int i = begin; i<end; i++) {
			l[i-begin] = letters[i];
		}
		return new Word(l);
	}
	
	protected boolean checkCentrals(Word w) {
		int i = 0;
		int j = 0;
		while(i<w.centrals) {
			while(j<centrals && letters[j].compareTo(w.letters[i])<0) {
				j++;
			}
			if(j==centrals || letters[j]!=w.letters[i]) {
				return false;
			}
			i++;
			j++;
		}
		return true;
	}
	
	public int find(Word w, int from) {
		if(!checkCentrals(w)) {
			return -1;
		}
		while(from<=(letters.length-centrals-w.letters.length+w.centrals)) {
			boolean eq = true;
			for(int i = 0; i<w.letters.length-w.centrals; i++) {
				if(letters[centrals+from+i]!=w.letters[w.centrals+i]) {
					eq = false;
					break;
				}
			}
			if(eq) {
				return from;
			}
			from++;
		}
		return -1;
	}
	
	public Word[] divide(Word w, int from) {
		int len1 = centrals-w.centrals+from;
		int len2 = letters.length-w.letters.length-len1;
		Letter[] arr1 = new Letter[len1];
		Letter[] arr2 = new Letter[len2];
		int j = 0;
		int pos = 0;
		for(int i = 0; i<centrals; i++) {
			if(j<w.centrals && w.letters[j]==letters[i]) {
				j++;
			} else {
				arr1[pos] = letters[i];
				pos++;
			}
		}
		for(int i = 0; i<from; i++) {
			arr1[pos+i] = letters[centrals+i];
		}
		int delta = letters.length-len2;
		for(int i = 0; i<len2; i++) {
			arr2[i] = letters[delta+i];
		}
		Word[] res = {new Word(arr1), new Word(arr2)};
		return res;
	}
	
	public static class Split {
		public Word left;
		public Word center;
		public Word right;
		public Split(Word left, Word center, Word right) {
			this.left = left;
			this.center = center;
			this.right = right;
		}
	}
	
	protected void splitCenters(Word w, LinkedList left, LinkedList center, LinkedList right) {
		int i = 0;
		int j = 0;
		while(i<centrals || j<w.centrals) {
			if(i<centrals && j<w.centrals && letters[i]==w.letters[j]) {
				center.add(letters[i]);
				i++;
				j++;
			} else if(j==w.centrals || (i<centrals && letters[i].compareTo(w.letters[j])<0)) {
				left.add(letters[i]);
				i++;
			} else {
				right.add(w.letters[j]);
				j++;
			}
		}
	}
	
	protected boolean checkSplit(Word w, int offset) {
		for(int i = 0; i+offset+centrals<letters.length; i++) {
			if(letters[i+offset+centrals]!=w.letters[i+w.centrals]) {
				return false;
			}
		}
		return true;
	}
	
	protected static void fillArray(Letter[] arr, Collection col) {
		int pos = 0;
		for(Iterator i = col.iterator(); i.hasNext();) {
			arr[pos] = (Letter) i.next();
			pos++;
		}
	}
	
	protected Split makeSplit(Word w, int offset, LinkedList cleft, LinkedList ccenter, LinkedList cright) {
		int centerSize = letters.length-centrals-offset;
		int len1 = offset+cleft.size();
		int len2 = centerSize+ccenter.size();
		int len3 = w.letters.length-w.centrals-centerSize+cright.size();
		Letter[] arr1 = new Letter[len1];
		Letter[] arr2 = new Letter[len2];
		Letter[] arr3 = new Letter[len3];
		fillArray(arr1, cleft);
		fillArray(arr2, ccenter);
		fillArray(arr3, cright);
		for(int i = 0; i<offset; i++) {
			arr1[cleft.size()+i] = letters[centrals+i];
		}
		for(int i = 0; i<centerSize; i++) {
			arr2[ccenter.size()+i] = letters[centrals+offset+i];
		}
		for(int i = 0; i<w.letters.length-w.centrals-centerSize; i++) {
			arr3[cright.size()+i] = w.letters[w.centrals+centerSize+i];
		}
		return new Split(new Word(arr1), new Word(arr2), new Word(arr3));
	}
	
	public Collection getAllSplits(Word w) {
		LinkedList splits = new LinkedList();
		LinkedList left = new LinkedList();
		LinkedList center = new LinkedList();
		LinkedList right = new LinkedList();
		splitCenters(w, left, center, right);
		int maxcenter = Math.min(letters.length-centrals, w.letters.length-w.centrals);
		int mincenter = 1;
		if(center.size()>0) {
			mincenter = 0;
		}
		for(int offset = letters.length-centrals-maxcenter; offset<=letters.length-centrals-mincenter; offset++) {
			if(checkSplit(w, offset)) {
				Split s = makeSplit(w, offset, left, center, right);
				if(s.left.getLength()+s.right.getLength()>0) {
					splits.add(s);
				}
			}
		}
		return splits;
	}
	
	public Letter[] getLetters() {
		return letters;
	}
	
	public static Word parseString(String s, Alphabet abc) {
		LinkedList letters = new LinkedList();
		int pos = 0;
		while(pos<s.length()) {
			while(pos<s.length() && (s.charAt(pos)==' ' || s.charAt(pos)=='*')) {
				pos++;
			}
			if(pos==s.length()) {
				break;
			}
			Letter l = abc.getFirstLetter(s.substring(pos));
			if(l==null) return null;
			pos+=l.getName().length();
			if(pos<s.length() && s.charAt(pos)=='^') {
				pos++;
				int pos1 = pos;
				while(pos<s.length() && s.charAt(pos)<='9' && s.charAt(pos)>='0') pos++;
				int power = Integer.parseInt(s.substring(pos1, pos));
				for(int i = 0; i<power; i++) {
					letters.add(l);
				}
			} else {
				letters.add(l);
			}
		}
		Letter[] arr = new Letter[letters.size()];
		Iterator i = letters.iterator();
		int j = 0;
		while(i.hasNext()) {
			arr[j] = (Letter) i.next();
			j++;
		}
		return new Word(arr);
	}
}
