package calculus;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import algebra.linear.Sum;
import algebra.numbers.Rational;
import algebra.words.Letter;
import algebra.words.Word;

/**
 * @author Anton Mellit
 *
 * To change this generated comment edit the template variable "typecomment":
 * Window>Preferences>Java>Templates.
 * To enable and disable the creation of type comments go to
 * Window>Preferences>Java>Code Generation.
 */
class Commutator {
	public WordSum x;
	public Sum values;
}

public class CenterAnalysis {
	protected Map comms;
	protected Letter[] letters;
	protected Letter[] generators;
	protected Set wordsToProcess;
	protected Set nextWords;
	protected Collection groebnerBase;
	protected boolean dump;
	protected int length;
	
	public CenterAnalysis(Letter[] generators, Letter[] letters, Collection groebnerBase) {
		comms = new TreeMap();
		this.letters = letters;
		this.generators = generators;
		wordsToProcess = new TreeSet();
		nextWords = new TreeSet();
		this.groebnerBase = groebnerBase;
		length = 0;
	}

	
	/**
	 * Sets the dump.
	 * @param dump The dump to set
	 */
	public void setDump(boolean dump) {
		this.dump = dump;
	}

	protected void init() {
		for(int i = 0; i<letters.length; i++) {
			if(!letters[i].isCentral()) {
				nextWords.add(new Word(letters[i]));
			}
		}
		length = 1;
	}
	
	protected WordSum process() {
		if(dump) {
			System.out.println("Processing words with length "+length+" ("+nextWords.size()+" words)");
		}
		wordsToProcess.clear();
		wordsToProcess.addAll(nextWords);
		nextWords.clear();
		for(Iterator i = wordsToProcess.iterator(); i.hasNext(); ) {
			Word w = (Word) i.next();
			if(isBaseWord(w)) {
//				System.out.println(w);
				addWords(w);
				Commutator comm = addCommutator(w);
				if(comm.values.isZero()) {
					if(dump) {
						System.out.println("Central element found: "+comm.x);
					}
					return comm.x;
				}
			}
		}
		length++;
		return null;
	}
	
	public WordSum findCenter() {
		init();
		WordSum answer = null;
		while(answer==null && !nextWords.isEmpty() && length<31) {
			answer = process();
		}
		return answer;
	}
	
	protected boolean isBaseWord(Word w) {
		for(Iterator i = groebnerBase.iterator(); i.hasNext(); ) {
			WordSum s = (WordSum) i.next();
			if(w.find((Word) s.getMainObject(), 0)>=0) {
				return false;
			}
		}
		return true;
	}
	
	protected void addWords(Word w) {
		for(int i = 0; i<letters.length; i++) {
//			if(length<9 || (!letters[i].isCentral()&&!w.getLetters()[0].isCentral())) {
				nextWords.add(w.multiply(new Word(letters[i]), Word.ONE));
//			}
		}
	}
	
	protected Commutator addCommutator(Word w) {
		Commutator comm = new Commutator();
		comm.values = new Sum();
		for(int i = 0; i<generators.length; i++) {
			addCommutator(comm.values, w, new Word(generators[i]));
		}
		comm.x = new WordSum();
		comm.x.add(w);
		if(!comm.values.isZero()) {
			reduce(comm);
		}
		if(!comm.values.isZero()) {
			comms.put(comm.values.getMainObject(), comm);
		}
		return comm;
	}
	
	protected void addCommutator(Sum s, Word w, Word x) {
		WordSum c = new WordSum();
		c.add(w.multiply(x, Word.ONE));
		c.add(w.multiply(Word.ONE, x), -1);
		c.reduce(groebnerBase);
		for(Iterator i = c.getObjects().iterator(); i.hasNext(); ) {
			Object o = i.next();
			s.add(x.toString()+": "+o, c.getKoeff(o));
		}
	}
	
	protected void reduce(Commutator comm) {
		boolean finished = false;
		do {
			Object main = comm.values.getMainObject();
			Commutator mod = (Commutator) comms.get(main);
			if(mod!=null) {
				Rational k = comm.values.getKoeff(main).divide(
					mod.values.getKoeff(mod.values.getMainObject())).multiply(-1);
				comm.values.add(mod.values, k);
				comm.x.add(mod.x, k);
			} else {
				finished = true;
			}
		} while(!finished && !comm.values.isZero());
	}
}
