package util;

import java.io.*;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import algebra.words.Alphabet;
import algebra.words.Letter;
import calculus.Groebner;
import calculus.WordSum;
/**
 * @author Anton Mellit
 *
 * To change this generated comment edit the template variable "typecomment":
 * Window>Preferences>Java>Templates.
 * To enable and disable the creation of type comments go to
 * Window>Preferences>Java>Code Generation.
 */
public class Setting {
	private Alphabet abc;
	private Groebner relations;
	private Vector generators;
	public Setting() {
		abc = new Alphabet();
		relations = new Groebner();
		generators = new Vector();
	}
	
	public void save(String file) throws IOException {
		PrintStream ps = new PrintStream(new FileOutputStream(file));
		ps.println("[Alphabet]");
		Letter[] letters = abc.getLetters();
		for (int j = 0; j < letters.length; j++) { //sort by index
			for (int i = 0; i < letters.length-1; i++) {
				if(letters[i].getIndex()>letters[i+1].getIndex()) {
					Letter l = letters[i]; letters[i] = letters[i+1]; letters[i+1] = l;
				}
			}
		}
		for (int i = 0; i < letters.length; i++) {
			ps.println("weight "+letters[i].getLevel()+(letters[i].isCentral()? " central ":" noncentral ")+
				letters[i].getName());
		}
		ps.println("[Generators]");
		for (Iterator iter = generators.iterator(); iter.hasNext();) {
			Letter l = (Letter) iter.next();
			ps.println(l);
		}
		ps.println("[Relations]");
		for (Iterator iter = relations.getRelations().iterator(); iter.hasNext();) {
			WordSum ws = (WordSum) iter.next();
			ps.println("\""+ws+"\"");
		}
		ps.close();
	}
	public void load(String file) throws IOException {
		InputStream is;
		StreamTokenizer s= new StreamTokenizer(new InputStreamReader(is=new FileInputStream(file)));
		s.parseNumbers();
		s.slashSlashComments(true);
		String section = "";
		s.eolIsSignificant(true);
		boolean commuting = false;
		int weight = 0;
		int k = 0;
		while(s.nextToken()!=s.TT_EOF) {
			if(s.ttype==s.TT_EOL) continue;
			if(s.ttype=='[') {
				if(s.nextToken()==s.TT_WORD) {
					section = s.sval;
					s.nextToken();
					continue;
				}
			}
			if(section.equalsIgnoreCase("Alphabet")) {
				if(s.ttype==s.TT_WORD) {
					if(s.sval.equalsIgnoreCase("Central")) {
						commuting = true;
					} else if(s.sval.equalsIgnoreCase("Noncentral")) {
						commuting = false;
					} else if(s.sval.equalsIgnoreCase("Weight")) {
						if(s.nextToken()==s.TT_NUMBER) {
							weight = (int) s.nval;
						} else {
							s.pushBack();
						}
					} else {	 
						k++;
						Letter l = new Letter(s.sval, commuting, weight, k);
						abc.addLetter(l);
					}
				}
			} else if(section.equalsIgnoreCase("Generators")) {
				if(s.ttype==s.TT_WORD) {
					generators.add(abc.getFirstLetter(s.sval));
				}
			} else if(section.equalsIgnoreCase("Relations")) {
				if(s.ttype=='"') {
					relations.addRelation(WordSum.parse(s.sval, abc));
				}
			}
		}
		is.close();
		relations.mergeRelations();
	}
	/**
	 * Returns the abc.
	 * @return Alphabet
	 */
	public Alphabet getAbc() {
		return abc;
	}

	/**
	 * Returns the relations.
	 * @return Groebner
	 */
	public Groebner getRelations() {
		return relations;
	}

	public Letter[] getGenerators() {
		Letter[] res = new Letter[generators.size()];
		return (Letter[]) generators.toArray(res);
	}
	
	public void dump() {
		abc.dump();		
		System.out.print("Generators:");
		Letter[] gens = getGenerators();
		for (int i = 0; i < gens.length; i++) {
			System.out.print(" "+gens[i]);
		}
		System.out.println();
		relations.dump();
	}
}
